#!/usr/bin/env python

import logging

#################################################################################
# Enable logging

log = logging.getLogger('test.bitbucket')

#################################################################################

import os
import sys
import pytest

from tests.helpers import GitRepoTestCase

from git_repo.services.service import bitbucket
from git_repo.exceptions import ResourceNotFoundError, ResourceError

class Test_BitBucket(GitRepoTestCase):
    log = log

    @property
    def local_namespace(self):
        if 'BITBUCKET_NAMESPACE' in os.environ:
            return os.environ['BITBUCKET_NAMESPACE']
        return 'git-repo-test'

    def get_service(self):
        service = bitbucket.BitbucketService(c=dict())
        return service

    def get_requests_session(self):
        return self.service.bb.client.session

    def test_00_fork(self):
        self.action_fork(local_namespace=self.local_namespace,
                        remote_namespace='abdo2015',
                        repository='git_tutorial')

    def test_01_create(self):
        self.action_create(namespace=self.local_namespace,
                           repository='foobar')


    def test_02_delete(self):
        self.action_delete(namespace=self.local_namespace,
                           repository='foobar')

    def test_03_delete_nouser(self):
        self.action_delete(repository='git_tutorial')

    def test_04_clone(self):
        self.action_clone(namespace='guyzmo',
                          repository='git-repo')

    def test_05_add(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo')

    def test_06_add__name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r')

    def test_07_add__alone(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True)

    def test_08_add__alone_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        alone=True)

    def test_09_add__default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        tracking='bitbucket')

    def test_10_add__default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        tracking='bitbucket')

    def test_11_add__alone_default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        tracking='bitbucket')

    def test_12_add__alone_default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        name='test0r',
                        tracking='bitbucket')

    def test_13_snippet_list(self):
        s_list = [
            ('https://bitbucket.org/snippets/guyzmo/ggd4X', 'test'),
            ('https://bitbucket.org/snippets/guyzmo/ggda9', 'test2'),
            ('https://bitbucket.org/snippets/guyzmo/77gaA', 'test3'),
        ]
        self.action_gist_list(gist_list_data=s_list)

    def test_14_snippet_list_with_gist(self):
        g_list = [
            'N.A.                  0 aoeu1.txt',
            'N.A.                  0 aeou2.txt',
        ]
        self.action_gist_list(gist='ggd4X')

    def test_15_snippet_list_with_bad_gist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_list(gist='42')

    def test_16_snippet_clone_with_gist(self):
        self.action_gist_clone(
            gist='https://bitbucket.org/snippets/{}/ggd4X'.format(os.environ.get('BITBUCKET_NAMESPACE', 'error')),
            clone_url='ssh://git@bitbucket.org/snippets/{}/ggd4X/test'.format(os.environ.get('BITBUCKET_NAMESPACE', 'error'))
        )

    def test_17_snippet_fetch_with_gist(self):
        content = self.action_gist_fetch(gist='77gaA', gist_file=None)
        assert content == '\n'.join([ 'ttt' ])

    def test_18_snippet_fetch_with_bad_gist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_fetch(gist='69', gist_file=None)

    def test_19_snippet_fetch_with_gist_file(self):
        content = self.action_gist_fetch(gist='ggd4X', gist_file='aoeu1.txt')
        assert content == '\n'.join([ '1' ])

    def test_20_snippet_fetch_with_bad_gist_file(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_fetch(gist='ggd4X', gist_file='failed')

    def test_21_snippet_create_gist_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a test.',
                gist_files=[ test_file ],
                secret=False)

    def test_22_snippet_create_gist_file_list(self, datadir):
        test_files = [
                str(datadir[ 'random-fortune-1.txt' ]),
                str(datadir[ 'random-fortune-2.txt' ]),
                str(datadir[ 'random-fortune-3.txt' ]),
                str(datadir[ 'random-fortune-4.txt' ]),
        ]
        self.action_gist_create(description='this is a test.',
                gist_files=test_files,
                secret=False)

    def test_23_snippet_create_gist_dir(self, datadir):
        test_dir = [
                str(datadir[ 'a_directory' ]),
        ]
        self.action_gist_create(description='this is a test.',
                gist_files=test_dir,
                secret=False)

    def test_24_snippet_create_gist_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a secret test.',
                gist_files=[ test_file ],
                secret=True)

    def test_25_snippet_create_gist_file_list(self, datadir):
        test_files = [
                str(datadir[ 'random-fortune-1.txt' ]),
                str(datadir[ 'random-fortune-2.txt' ]),
                str(datadir[ 'random-fortune-3.txt' ]),
                str(datadir[ 'random-fortune-4.txt' ]),
        ]
        self.action_gist_create(description='this is a secret test.',
                gist_files=test_files,
                secret=True)

    def test_26_snippet_create_gist_dir(self, datadir):
        test_dir = [
                str(datadir[ 'a_directory' ]),
        ]
        self.action_gist_create(description='this is a secret test.',
                gist_files=test_dir,
                secret=True)

    def test_27_snippet_delete(self):
        self.action_gist_delete(gist='p4RzX')

    def test_28_snippet_delete__not_exist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_delete(gist='p4RzX')

    def test_29_snippet_create_snippet__file_not_exist(self, datadir):
        with pytest.raises(FileNotFoundError):
            test_dir = [
                    'does_not_exists'
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_dir,
                    secret=False)

    def test_30_request_list(self):
        self.action_request_list(
                namespace='atlassian',
                repository='python-bitbucket',
                rq_list_data=[
            ('1', 'Bugfixes for two crashes appearing during reply parsing of repository_create (output of this command is different from normal status, unfortunately). The latter appeared after I fixed the former.', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/1'),
            ('2', 'Attempt to fix bitbucket_create crash again, this time py3k compatible', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/2'),
            ('3', '[WIP] Initial support for webhooks', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/3'),
            ('4', 'fix: lookups for PRs based on state were calling a non existant method', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/4'),
            ('5', '🚧 Fork feature implementation', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/5'),
            ('6', '💄 Updated code style to be closer to python3 style', 'https://bitbucket.org/atlassian/python-bitbucket/pull-requests/6'),
        ])

    def test_30_request_list_empty(self):
        self.action_request_list(
                namespace='guyzmo',
                repository='git-repo',
                rq_list_data=[ ])

    def test_31_request_fetch__good(self):
        namespace='atlassian'
        repository='python-bitbucket'
        request='2'
        local_branch='requests/bitbucket'
        local_slug = self.service.format_path(namespace=namespace, repository=repository, rw=False)
        remote_branch='repo_create_bugfix_again'
        remote_name='requests-bitbucket-Mekk-python-bitbucket'
        with self.recorder.use_cassette(self._make_cassette_name(1)):
            with self.mockup_git(namespace, repository):
                self.set_mock_popen_commands([
                    ('git remote add all {}'.format(local_slug), b'', b'', 0),
                    ('git remote add {} {}'.format(self.service.name, local_slug), b'', b'', 0),
                    ('git remote add requests-bitbucket-Mekk-python-bitbucket https://bitbucket.org/Mekk/python-bitbucket', b'', b'', 0),
                    ('git version', b'git version 2.8.0', b'', 0),
                    ('git pull --progress -v {} master'.format(self.service.name), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * branch            master     -> FETCH_HEAD',
                        ' * [new branch]      master     -> {1}/{0}'.format(request, local_branch)]).encode('utf-8'),
                    0),
                    ('git fetch --progress -v {0} {2}/{1}/head:{3}/{1}'.format(
                            self.service.name,
                            request,
                            remote_branch,
                            local_branch), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * [new branch]      master     -> request/{}'.format(request)]).encode('utf-8'),
                    0),
                    ('git fetch --progress -v {} {}:{}/{}'.format(
                            remote_name,
                            remote_branch,
                            local_branch, request), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * [new branch]      {}     -> {}/{}'.format(remote_branch, local_branch, request),
                        ' * [new branch]      {}     -> {}/{}'.format(remote_branch, remote_name, remote_branch),
                        ]).encode('utf-8'),
                    0)
                    ])
                self.service.connect()
                self.service.clone(namespace, repository, rw=False)
                self.service.repository.create_remote('all', url=local_slug)
                self.service.repository.create_remote(self.service.name, url=local_slug)
                self.service.request_fetch(repo=repository, user=namespace, request=request)

    @pytest.mark.skip
    # Skipping this test because of bug in gitpython, cf https://github.com/gitpython-developers/gitpython#:
    #   > Pumping 'stderr' of cmd(['git', 'pull', '--progress', '-v', 'bitbucket', 'master']) failed due to: TypeError("a bytes-like object is required, not 'str'",)
    #   Exception in thread Thread-10:
    #   Traceback (most recent call last):
    #     File "/home/travis/virtualenv/python3.6.1/lib/python3.6/site-packages/git/cmd.py", line 87, in pump_stream
    #       handler(line)
    #     File "/home/travis/virtualenv/python3.6.1/lib/python3.6/site-packages/git/util.py", line 483, in handler
    #       return self._parse_progress_line(line.rstrip())
    #     File "/home/travis/virtualenv/python3.6.1/lib/python3.6/site-packages/git/util.py", line 388, in _parse_progress_line
    #       if len(self.error_lines) > 0 or self._cur_line.startswith(('error:', 'fatal:')):
    #   TypeError: a bytes-like object is required, not 'str'
    #   During handling of the above exception, another exception occurred:
    #   Traceback (most recent call last):
    #     File "/opt/python/3.6.1/lib/python3.6/threading.py", line 916, in _bootstrap_inner
    #       self.run()
    #     File "/opt/python/3.6.1/lib/python3.6/threading.py", line 864, in run
    #       self._target(*self._args, **self._kwargs)
    #     File "/home/travis/virtualenv/python3.6.1/lib/python3.6/site-packages/git/cmd.py", line 90, in pump_stream
    #       raise CommandError(['<%s-pump>' % name] + cmdline, ex)
    #   git.exc.CommandError: Cmd('<stderr-pump>') failed due to: TypeError('a bytes-like object is required, not 'str'')
    #     cmdline: <stderr-pump> git pull --progress -v bitbucket master
    def test_31_request_fetch__bad_request(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_fetch(namespace='atlassian',
                repository='python-bitbucket',
                request='42',
                fail=True)

    def test_32_request_create__good(self):
        r = self.action_request_create(namespace=self.local_namespace,
                repository='test_create_requests',
                target_branch='master',
                source_branch='pr-test',
                title='PR test',
                description='PR description')
        assert r == {
                'local': 'pr-test',
                'ref': 1,
                'remote': 'master',
                'project': '{}/test_create_requests'.format(self.local_namespace),
                'url': 'https://bitbucket.org/{}/test_create_requests/pull-requests/1'.format(self.local_namespace),
                }

    def test_32_request_create__bad_branch(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_create(namespace=self.local_namespace,
                    repository='test_create_requests',
                    target_branch='master',
                    source_branch='does_not_exists',
                    title='PR test',
                    description='PR description')

    def test_32_request_create__bad_repo(self):
        with pytest.raises(ResourceNotFoundError):
            r = self.action_request_create(namespace=self.local_namespace,
                    repository='does_not_exists',
                    target_branch='master',
                    source_branch='pr-test',
                    title='PR test',
                    description='PR description')

    @pytest.mark.skip
    # will need cassette regeneration
    def test_32_request_create__guess_branch(self):
        r = self.action_request_create(namespace=self.local_namespace,
                repository='test_create_requests',
                target_branch=None,
                source_branch=None,
                title='PR test',
                description='PR description')
        assert r == {
                'local': 'pr-test',
                'ref': 1,
                'remote': 'master',
                'project': '{}/test_create_requests'.format(self.local_namespace),
                'url': 'https://bitbucket.org/{}/test_create_requests/pull-requests/1'.format(self.local_namespace),
                }

    def test_33_open(self):
        self.action_open(namespace='guyzmo',
                         repository='git-repo')

    def test_34_list__short(self, caplog):
        projects = self.action_list(namespace='git-repo-test')
        assert projects == ['{}', ('Total repositories: 1',), ['git-repo-test/git-repo']]
        assert 'GET /2.0/users/git-repo-test' in caplog.text
        assert 'GET /2.0/repositories/git-repo-test HTTP/1.1' in caplog.text

    def test_34_list__long(self, caplog):
        projects = self.action_list(namespace='git-repo-test', _long=True)
        assert projects == ['{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}',
                ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif', 'Name'],
                ['F ', '92', '1', 'N.A.', '1', 'N.A.', '1', 'N.A.', 'python', '2016-03-30T13:30:15.637449+00:00', 'git-repo-test/git-repo']]
        assert 'GET /2.0/users/git-repo-test' in caplog.text
        assert 'GET /2.0/repositories/git-repo-test HTTP/1.1' in caplog.text





#!/usr/bin/env python

import os
import sys
import logging

import pytest

from unittest.mock import mock_open, patch

#################################################################################
# Enable logging

log = logging.getLogger('test.github')

#################################################################################

from tests.helpers import GitRepoTestCase

from git_repo.services.service import github
from git_repo.exceptions import ResourceExistsError, ResourceNotFoundError, ResourceError


class Test_Github(GitRepoTestCase):
    log = log
    namespace = os.environ['GITHUB_NAMESPACE']

    @property
    def local_namespace(self):
        if 'GITHUB_NAMESPACE' in os.environ:
            return os.environ['GITHUB_NAMESPACE']
        return 'git-repo-test'

    def get_service(self):
        return github.GithubService(c=dict())

    def get_requests_session(self):
        return self.service.gh._session

    def test_00_fork(self):
        self.action_fork(local_namespace=self.local_namespace,
                         remote_namespace='sigmavirus24',
                         repository='github3.py')

    def test_01_create__new(self):
        self.action_create(namespace=self.local_namespace,
                           repository='foobar')

    def test_01_create__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace=self.local_namespace,
                            repository='git-repo')

    def test_01_create_organization__new(self):
        self.action_create(namespace='test-git-repo',
                           repository='foobar')

    def test_01_create_organization__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace='test-git-repo',
                            repository='git-repo')


    def test_02_delete(self):
        self.action_delete(namespace=self.local_namespace,
                           repository='foobar')

    def test_03_delete_nouser(self):
        self.action_delete(repository='github3.py')

    def test_04_clone(self):
        self.action_clone(namespace='git-services',
                          repository='github-testing')

    def test_05_add(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo')

    def test_06_add__name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r')

    def test_07_add__alone(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True)

    def test_08_add__alone_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        alone=True)

    def test_09_add__default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        tracking='github')

    def test_10_add__default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        tracking='github')

    def test_11_add__alone_default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        tracking='github')

    def test_12_add__alone_default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        name='test0r',
                        tracking='github')

    def test_12_add__upstream(self):
        self.action_add(namespace=None,
                        repository='upstream',
                        alone=True,
                        tracking=False,
                        auto_slug=True,
                        remotes={'github': 'https://github.com/guyzmo/github-testing'})

    @pytest.mark.skip
    def test_12_add__guess(self):
        mock_open(read_data='1\n')
        with patch('sys.stdin', mock_open):
            self.action_add(namespace=self.local_namespace,
                            repository='git-repo',
                            alone=True,
                            tracking='github',
                            auto_slug=True,
                            remotes={
                                'origin': 'https://github.com/foo/bar',
                                'wootwoot': 'git@github.com:w00t/w00t',
                                'duckling': 'ssh://github.com/duck/duck',
                            }
            )

    def test_13_gist_list(self):
        g_list = [
            '{:45.45} {}',
            ("title", "url"),
            ("unicode combined class for better character counting and indexing",           "https://gist.github.com/a7ce4fddba7744ddf335" ) ,
            ("avr-gcc derivation build",                                                    "https://gist.github.com/7e5a12bc158a79966020" ) ,
            ("avr-gcc derivation",                                                          "https://gist.github.com/dd9ab22c8f22a5a8f3d1" ) ,
            ("`brew cask install mplabx` fixed",                                            "https://gist.github.com/893fbc98bf1c9cf6212a" ) ,
            ("`brew cask install mplabx` issue with target in /usr/local/mplabx",           "https://gist.github.com/9baed8712a16a29c2e90" ) ,
            ("cask formula for mplabx (stripped down)",                                     "https://gist.github.com/1c03ddfdc8f57fa7919a" ) ,
            ("`brew cask install mplabx` issue removing the caskroom dir!",                 "https://gist.github.com/21949a3fa5f981a869bf" ) ,
            ("cask formula for mplabx",                                                     "https://gist.github.com/588deedefc1675998bbe" ) ,
            ("TaskJuggler example",                                                         "https://gist.github.com/281502e4ae6fce01db92" ) ,
            ("redmine to taskjuggler recursive func",                                       "https://gist.github.com/06c5b0da8d10c514166f" ) ,
            ("Simple example of Flask/Presst with Login and Principals (not working!)",     "https://gist.github.com/603ccdd0f504c63cd0df" ) ,
            ("radial colour picker crash",                                                  "https://gist.github.com/2482962b153ebd5cfa6b" ) ,
            ("Platform levelling GCode file",                                               "https://gist.github.com/fc46896f3e604269ff93" ) ,
            ("Code to download movies from pluzz",                                          "https://gist.github.com/3b18193d6bea07bac37c" ) ,
            ("",                                                                            "https://gist.github.com/c01613d0453df275622a" ) ,
            ("I2C scanner code for SL030",                                                  "https://gist.github.com/10118958"             ) ,
            ("",                                                                            "https://gist.github.com/5b79437ddd3f49491ce3" ) ,
            ("An enhanced implementation of authenticating to stackoverflow using python.", "https://gist.github.com/5730750"              ) ,
            ("Patch to be applied for homebrew's strigi.rb Formula.",                       "https://gist.github.com/4308707"              ) ,
            ("Patch for strnlen support in freevpn (for OSX 10.6)",                         "https://gist.github.com/4170462"              ) ,
            ("z.sh patch for multiline quotes",                                             "https://gist.github.com/3666086"              ) ,
            ("A Tornado-based library that enables Event Source support !",                 "https://gist.github.com/2788003"              ) ,
        ]
        self.action_gist_list(gist_list_data=g_list)

    def test_14_gist_list_with_gist(self):
        g_list = [
            "C                 2542  i2c_scanner.c"
        ]
        self.action_gist_list(gist='10118958')

    def test_15_gist_list_with_bad_gist(self):
        self.action_gist_list(gist='42')

    def test_16_gist_clone_with_gist(self):
        self.action_gist_clone(gist='https://gist.github.com/10118958')

    def test_17_gist_fetch_with_gist(self):
        content = self.action_gist_fetch(gist='4170462', gist_file=None)
        assert content == '\n'.join([
            'diff --git a/platform/io.c b/platform/io.c',
            'index 209666a..0a6c2cf 100644',
            '--- a/platform/io.c',
            '+++ b/platform/io.c',
            '@@ -24,6 +24,16 @@',
            ' #if defined(__FreeBSD__)',
            ' #define IO_BSD',
            ' #elif defined(__APPLE__)',
            '+size_t strnlen(const char *s, size_t maxlen) ',
            '+{ ',
            '+        size_t len; ',
            '+ ',
            '+        for (len = 0; len < maxlen; len++, s++) { ',
            '+                if (!*s) ',
            '+                        break; ',
            '+        } ',
            '+        return (len); ',
            '+} ',
            ' #define IO_BSD',
            ' #define IO_USE_SELECT',
            ' #elif defined(WIN32)',
        ])

    def test_18_gist_fetch_with_bad_gist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_fetch(gist='69', gist_file=None)

    def test_19_gist_fetch_with_gist_file(self):
        content = self.action_gist_fetch(gist='4170462', gist_file='freevpn0.029__platform__io.patch')
        assert content == '\n'.join([
            'diff --git a/platform/io.c b/platform/io.c',
            'index 209666a..0a6c2cf 100644',
            '--- a/platform/io.c',
            '+++ b/platform/io.c',
            '@@ -24,6 +24,16 @@',
            ' #if defined(__FreeBSD__)',
            ' #define IO_BSD',
            ' #elif defined(__APPLE__)',
            '+size_t strnlen(const char *s, size_t maxlen) ',
            '+{ ',
            '+        size_t len; ',
            '+ ',
            '+        for (len = 0; len < maxlen; len++, s++) { ',
            '+                if (!*s) ',
            '+                        break; ',
            '+        } ',
            '+        return (len); ',
            '+} ',
            ' #define IO_BSD',
            ' #define IO_USE_SELECT',
            ' #elif defined(WIN32)',
        ])

    def test_20_gist_fetch_with_bad_gist_file(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_fetch(gist='4170462', gist_file='failed')

    def test_21_gist_create_gist_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a test.',
                gist_files=[ test_file ],
                secret=False)

    def test_22_gist_create_gist_file_list(self, datadir):
        test_files = [
                str(datadir[ 'random-fortune-1.txt' ]),
                str(datadir[ 'random-fortune-2.txt' ]),
                str(datadir[ 'random-fortune-3.txt' ]),
                str(datadir[ 'random-fortune-4.txt' ]),
        ]
        self.action_gist_create(description='this is a test.',
                gist_files=test_files,
                secret=False)

    def test_23_gist_create_gist_dir(self, datadir):
        test_dir = [
                str(datadir[ 'a_directory' ]),
        ]
        self.action_gist_create(description='this is a test.',
                gist_files=test_dir,
                secret=False)

    def test_24_gist_create_gist_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a secret test.',
                gist_files=[ test_file ],
                secret=True)

    def test_25_gist_create_gist_file_list(self, datadir):
        test_files = [
                str(datadir[ 'random-fortune-1.txt' ]),
                str(datadir[ 'random-fortune-2.txt' ]),
                str(datadir[ 'random-fortune-3.txt' ]),
                str(datadir[ 'random-fortune-4.txt' ]),
        ]
        self.action_gist_create(description='this is a secret test.',
                gist_files=test_files,
                secret=True)

    def test_26_gist_create_gist_dir(self, datadir):
        test_dir = [
                str(datadir[ 'a_directory' ]),
        ]
        self.action_gist_create(description='this is a secret test.',
                gist_files=test_dir,
                secret=True)

    def test_27_gist_delete(self):
        self.action_gist_delete(gist='7dcc495dda5e684cba94940a01f60e95')

    def test_28_gist_delete__not_exist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_delete(gist='7dcc495dda5e684cba94940a01f60e95')

    def test_29_gist_create_gist__file_not_exist(self, datadir):
        with pytest.raises(FileNotFoundError):
            test_dir = [
                    'does_not_exists'
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_dir,
                    secret=False)

    def test_30_request_list(self):
        self.action_request_list(
                namespace='guyzmo',
                repository='git-repo',
                rq_list_data=[
            '{}\t{:<60}\t{}',
            ('id', 'title', 'URL'),
            ('3', 'docs for fqdn > url', 'https://github.com/guyzmo/git-repo/pull/3'),
            ('2', 'prefer gitrepo.<target>.token > privatekey, docs', 'https://github.com/guyzmo/git-repo/pull/2'),
        ])

    def test_31_request_fetch(self):
        self.action_request_fetch(
                namespace='git-services',
                repository='github-testing',
                request='2',
                remote_branch='pull',
                local_branch='requests/github')

    def test_31_request_fetch__bad_request(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_fetch(
                namespace='git-services',
                repository='github-testing',
                request='1',
                remote_branch='pull',
                local_branch='requests/github',
                fail=True)

    def test_32_request_create(self):
        r = self.action_request_create(namespace=self.namespace,
                repository='test_create_requests',
                source_branch='pr-test',
                target_branch='master',
                title='PR test',
                description='PR description')
        assert r == {
                'local': 'pr-test',
                'ref': 1,
                'remote': 'master',
                'project': '{}/test_create_requests'.format(self.namespace),
                'url': 'https://github.com/{}/test_create_requests/pull/1'.format(self.namespace),
                }

    def test_32_request_create__bad_branch(self):
        with pytest.raises(ResourceError):
            self.action_request_create(namespace=self.namespace,
                    repository='test_create_requests',
                    source_branch='does_not_exists',
                    target_branch='master',
                    title='PR test',
                    description='PR description')

    def test_32_request_create__bad_repo(self):
        with pytest.raises(ResourceNotFoundError):
            r = self.action_request_create(namespace=self.namespace,
                    repository='does_not_exists',
                    source_branch='pr-test',
                    target_branch='master',
                    title='PR test',
                    description='PR description')

    def test_32_request_create__guess_branch(self):
        r = self.action_request_create(namespace=self.namespace,
                repository='test_create_requests',
                source_branch=None,
                target_branch=None,
                title='PR test',
                description='PR description')

    def test_33_open(self):
        self.action_open(namespace='guyzmo',
                         repository='git-repo')

    def test_34_list__short(self, caplog):
        projects = self.action_list(namespace='git-repo-test')
        assert projects == ['{}', ('Total repositories: 1',), ['git-repo-test/git-repo']]
        assert 'GET https://api.github.com/users/git-repo-test/repos' in caplog.text

    def test_34_list__long(self, caplog):
        projects = self.action_list(namespace='git-repo-test', _long=True)
        assert projects == ['{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}',
                ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif', 'Name'],
                ['F ', '92', '0', '0', '0', '1', '0', '0', 'Python', 'Mar 30 2016', 'git-repo-test/git-repo']]
        assert 'GET https://api.github.com/users/git-repo-test/repos' in caplog.text





#!/usr/bin/env python

import logging

#################################################################################
# Enable logging

log = logging.getLogger('test.gitlab')

#################################################################################

import os
import sys
import pytest

from tests.helpers import GitRepoTestCase

from git_repo.services.service import gitlab
from git_repo.exceptions import ResourceExistsError, ResourceNotFoundError, ResourceError, ArgumentError

class Test_Gitlab(GitRepoTestCase):
    log = log
    namespace = os.environ['GITLAB_NAMESPACE']

    @property
    def local_namespace(self):
        if 'GITLAB_NAMESPACE' in os.environ:
            return os.environ['GITLAB_NAMESPACE']
        return 'git-repo-test'

    def get_service(self):
        return gitlab.GitlabService(c=dict())

    def get_requests_session(self):
        return self.service.gl.session

    def test_00_fork(self):
        self.action_fork(local_namespace=self.local_namespace,
                         remote_namespace='sigmavirus24',
                         repository='github3-py')

    def test_01_create__new(self):
        self.action_create(namespace=self.local_namespace,
                           repository='foobar')

    def test_01_create__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace=self.local_namespace,
                            repository='git-repo')

    def test_01_create_group__new(self):
        self.action_create(namespace='git-repo-test',
                           repository='foobar')

    def test_01_create_group__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace='git-repo-test',
                            repository='git-repo')


    def test_02_delete(self):
        self.action_delete(namespace=self.local_namespace,
                           repository='foobar')

    def test_03_delete_nouser(self):
        self.action_delete(repository='github3-py')

    def test_04_clone(self):
        self.action_clone(namespace='guyzmo',
                          repository='git-repo')

    def test_05_add(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo')

    def test_06_add__name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r')

    def test_07_add__alone(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True)

    def test_08_add__alone_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        alone=True)

    def test_09_add__default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        tracking='gitlab')

    def test_10_add__default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        name='test0r',
                        tracking='gitlab')

    def test_11_add__alone_default(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        tracking='gitlab')

    def test_12_add__alone_default_name(self):
        self.action_add(namespace='guyzmo',
                        repository='git-repo',
                        alone=True,
                        name='test0r',
                        tracking='gitlab')

    def test_13_snippet_list_alone(self):
        namespace = os.environ.get('GITLAB_NAMESPACE', 'bogus')
        s_list = [
            '{:45.45} {}',
            ("title", "url"),
            ('this is a secret test.', 'https://gitlab.com/snippets/34124',                               ) ,
            ('this is a test.',        'https://gitlab.com/snippets/34121',                               ) ,
            ('this is a test.',        'https://gitlab.com/{}/git-repo/snippets/32318'.format(namespace), ) ,
            ('this is a secret test.', 'https://gitlab.com/{}/git-repo/snippets/32317'.format(namespace), ) ,
            ('this is a test.',        'https://gitlab.com/{}/git-repo/snippets/32316'.format(namespace), ) ,
            ('requirements.txt',       'https://gitlab.com/{}/git-repo/snippets/32303'.format(namespace), ) ,
            ('test',                   'https://gitlab.com/{}/git-repo/snippets/26173'.format(namespace), ) ,
            ('test',                   'https://gitlab.com/snippets/20696',                               )
        ]
        self.action_gist_list(gist_list_data=s_list)

    def test_13_snippet_list_with_project(self):
        self.action_gist_list(gist='guyzmo/git-repo')

    def test_13_snippet_list_with_non_existent_project(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_list(gist='non-existent')

    def test_14_snippet_clone(self):
        with pytest.raises(ArgumentError):
            self.action_gist_clone(gist='XXX')

    def test_15_snippet_fetch_global_snippet(self):
        snippet = self.action_gist_fetch(gist='20696')
        assert snippet == "test"

    def test_15_snippet_fetch_project_snippet(self):
        content = self.action_gist_fetch(gist='https://gitlab.com/guyzmo/git-repo/snippets/32303')
        assert content == '\n'.join([
            'docopt',
            'progress',
            'GitPython>=2.1.0',
            'uritemplate.py==2.0.0',
            'github3.py==0.9.5',
            'python-gitlab>=0.13',
            'bitbucket-api', ''
        ])

    def test_15_snippet_fetch_with_bad_project_snippet(self):
        with pytest.raises(ResourceNotFoundError):
            content = self.action_gist_fetch(gist='https://gitlab.com/guyzmo/git-repo/snippets/42')

    def test_16_snippet_create_snippet_global_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a test.',
                gist_files=[ test_file ],
                secret=False)

    def test_16_snippet_create_snippet_project_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a test.',
                gist_files=[ 'guyzmo/git-repo', test_file ],
                secret=False)

    def test_16_snippet_create_snippet_file_list(self, datadir):
        with pytest.raises(ArgumentError):
            test_files = [
                    str(datadir[ 'random-fortune-1.txt' ]),
                    str(datadir[ 'random-fortune-2.txt' ]),
                    str(datadir[ 'random-fortune-3.txt' ]),
                    str(datadir[ 'random-fortune-4.txt' ]),
            ]
            self.action_gist_create(description='this is a test.',
                    gist_files=test_files,
                    secret=False)

    def test_16_snippet_create_snippet_dir(self, datadir):
        with pytest.raises(IsADirectoryError):
            test_dir = [
                    str(datadir[ 'a_directory' ]),
            ]
            self.action_gist_create(description='this is a test.',
                    gist_files=test_dir,
                    secret=False)

    def test_16_snippet_create_secret_snippet_global_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a secret test.',
                gist_files=[ test_file ],
                secret=True)

    def test_16_snippet_create_secret_snippet_project_file(self, datadir):
        test_file = str(datadir[ 'random-fortune-1.txt' ])
        self.action_gist_create(description='this is a secret test.',
                gist_files=[ 'git-repo', test_file ],
                secret=True)

    def test_16_snippet_create_snippet__file_not_exist(self, datadir):
        with pytest.raises(FileNotFoundError):
            test_dir = [
                    'does_not_exists'
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_dir,
                    secret=False)

    def test_17_snippet_project_delete(self):
        self.action_gist_delete(gist='https://gitlab.com/guyzmo/git-repo/snippets/32303')

    def test_17_snippet_project_delete__not_exist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_delete(gist='https://gitlab.com/guyzmo/git-repo/snippets/32303')

    def test_17_snippet_global_delete(self):
        self.action_gist_delete(gist='https://gitlab.com/snippets/34111')

    def test_17_snippet_global_delete__not_exist(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_gist_delete(gist='https://gitlab.com/snippets/32304')

    def test_18_request_list(self):
        self.action_request_list(
                namespace='git-repo-test',
                repository='git-repo',
                rq_list_data=[
            '{:>3}\t{:<60}\t{:2}',
            ('id', 'title', 'URL'),
            ('1', 'Adding gitlab gists and requests feature', 'https://gitlab.com/git-repo-test/git-repo/merge_requests/1'),
        ])

    def test_19_request_fetch(self):
        self.action_request_fetch(namespace=self.local_namespace,
                repository='git-repo',
                request='4',
                remote_branch='merge-requests',
                local_branch='requests/gitlab')

    def test_19_request_fetch__bad_request(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_fetch(namespace=self.local_namespace,
                repository='git-repo',
                request='42',
                remote_branch='merge_requests',
                local_branch='requests/gitlab',
                fail=True)

    def test_20_request_create(self):
        r = self.action_request_create(namespace=self.local_namespace,
                repository='test_create_requests',
                source_branch='pr-test',
                target_branch='master',
                title='PR test',
                description='PR description')
        assert r == {
            'local': 'pr-test',
            'project': '{}/test_create_requests'.format(self.local_namespace),
            'ref': 1,
            'remote': 'master',
            'url': 'https://gitlab.com/{}/test_create_requests/merge_requests/1'.format(self.local_namespace)
        }

    def test_20_request_create__bad_branch(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_create(namespace=self.local_namespace,
                    repository='test_create_requests',
                    source_branch='this_is_not_a_branch',
                    title='PR test',
                    description='PR description')

    def test_20_request_create__bad_repo(self):
        with pytest.raises(ResourceNotFoundError):
            r = self.action_request_create(namespace=self.local_namespace,
                    repository='does_not_exists',
                    source_branch='pr-test',
                    target_branch='master',
                    title='PR test',
                    description='PR description')

    def test_20_request_create__blank_branch(self):
        with pytest.raises(ResourceError):
            r = self.action_request_create(namespace=self.local_namespace,
                    repository='test_create_requests',
                    source_branch=None,
                    target_branch=None,
                    title='PR test',
                    description='PR description')

    def test_31_open(self):
        self.action_open(namespace=self.local_namespace,
                         repository='git-repo')




#!/usr/bin/env python

from tempfile import TemporaryDirectory
from git import Repo, Git

from testfixtures import Replace, ShouldRaise, compare
from testfixtures.popen import MockPopen

from contextlib import contextmanager

import os
import sys
import logging
import betamax

from git_repo.repo import RepositoryService, main


class TestGitPopenMockupMixin:
    def setup_git_popen(self):
        # repository mockup (in a temporary place)
        self.repository = Repo.init(self.tempdir.name)
        # setup git command mockup
        self.Popen = MockPopen()
        self.Popen.mock.Popen_instance.stdin = None
        self.Popen.mock.Popen_instance.wait = lambda *a, **k: self.Popen.wait()
        self.Popen.mock.Popen_instance.__enter__ = lambda self: self
        self.Popen.mock.Popen_instance.__exit__ = lambda self, *a, **k: None

    def set_mock_popen_commands(self, cmd_list):
        for cmd, out, err, rc in cmd_list:
            self.Popen.set_command(cmd, out, err, returncode=rc)

    def mockup_git(self, namespace, repository, url=None):
        # disable refspec check
        from git import remote
        remote.Remote._assert_refspec = lambda self: None
        # write FETCH_HEAD ref
        with open(os.path.join(self.repository.git_dir, 'FETCH_HEAD'), 'w') as f:
            url = url or "{}:{}/{}".format(self.service.fqdn, namespace, repository)
            f.write("749656b8b3b282d11a4221bb84e48291ca23ecc6" \
                    "		branch 'master' of {}".format(url))
        return Replace('git.cmd.Popen', self.Popen)


class RepositoryMockup(RepositoryService):
    name = 'test_name'
    command = 'test_command'
    fqdn = 'example.org'
    def __init__(self, *args, **kwarg):
        super(RepositoryMockup, self).__init__(*args, **kwarg)
        self._did_pull = None
        self._did_clone = None
        self._did_add = None
        self._did_open = None
        self._did_connect = None
        self._did_delete = None
        self._did_create = None
        self._did_fork = None
        self._did_user = False
        self._did_gist_list = None
        self._did_gist_fetch = None
        self._did_gist_clone = None
        self._did_gist_create = None
        self._did_gist_delete = None
        self._did_request_create = None
        self._did_request_list = None
        self._did_request_fetch = None

    def pull(self, *args, **kwarg):
        self._did_pull = (args, kwarg)

    def clone(self, *args, **kwarg):
        self._did_clone = (args, kwarg)

    def add(self, repo, user=None, *args, **kwarg):
        self._did_add = ((user, repo)+args, kwarg)
        class FakeRemote:
            name = 'foobar'
        return FakeRemote, user, repo

    def open(self, *args, **kwarg):
        self._did_open = (args, kwarg)

    def connect(self):
        self._did_connect = True

    def delete(self, *args, **kwarg):
        self._did_delete = (tuple(reversed(args)), kwarg)

    def create(self, *args, **kwarg):
        self._did_create = (args, kwarg)

    def fork(self, *args, **kwarg):
        self._did_fork = (args, kwarg)

    def gist_list(self, *args, **kwarg):
        self._did_gist_list = (args, kwarg)
        if len(args) == 0 or not args[0]:
            yield '{} {}'
            yield 'title', 'url'
            yield 'id1', 'value1'
            yield 'id2', 'value2'
            yield 'id3', 'value3'
        elif len(args) == 1:
            if args[0] == 'bad':
                raise Exception('bad gist!')
            else:
                yield '{} {} {}'
                yield 'language', 'size', 'name'
                yield 'lang1', 'size1', 'name1'
                yield 'lang2', 'size2', 'name2'
                yield 'lang3', 'size3', 'name3'

    def gist_fetch(self, *args, **kwarg):
        self._did_gist_fetch = (args, kwarg)
        if args[0] == 'bad':
            raise Exception('bad gist!')
        elif args[1] == 'bad':
            raise Exception('bad gist file!')
        else:
            return "content of a gist"

    def gist_clone(self, *args, **kwarg):
        self._did_gist_clone = (args, kwarg)
        if args[0] == 'bad':
            raise Exception('bad gist!')

    def gist_create(self, *args, **kwarg):
        self._did_gist_create = (args, kwarg)
        if 'exists' in args[0]:
            raise Exception('gist exists!')
        return 'https://gists/42'

    def gist_delete(self, *args, **kwarg):
        self._did_gist_delete = (args, kwarg)
        if args[0] == 'bad':
            raise Exception('bad gist!')

    def request_list(self, *args, **kwarg):
        self._did_request_list = (args, kwarg)
        yield '{} {} {}'
        yield ('id', 'description', 'URL')
        yield ('1', 'desc1', 'http://request/1')
        yield ('2', 'desc2', 'http://request/2')
        yield ('3', 'desc3', 'http://request/3')

    def request_fetch(self, *args, **kwarg):
        self._did_request_fetch = (args, kwarg)
        if args[-1] == 'bad':
            raise Exception('bad request for merge!')
        return "pr/42"

    def request_create(self, *args, **kwarg):
        self._did_request_create = (args, kwarg)
        if args[2] == 'bad' or args[3] == 'bad':
            raise Exception('bad branch to request!')
        local = args[2] or 'pr-test'
        remote = args[3] or 'base-test'
        return {'local': local, 'remote': remote, 'project': '/'.join(args[:2]), 'ref': 42}

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        return '{}:{}'.format(login, password)

    @property
    def user(self):
        self._did_user = True
        return 'foobar'

    def get_repository(self, *args, **kwarg):
        return {}


class GitRepoMainTestCase(TestGitPopenMockupMixin):
    def setup_method(self, method):
        self.log.info('GitRepoMainTestCase.setup_method({})'.format(method))
        self.tempdir = TemporaryDirectory()
        RepositoryService.service_map = {
            'github': RepositoryMockup,
            'gitlab': RepositoryMockup,
            'bitbucket': RepositoryMockup,
        }
        RepositoryService.command_map = {
            'hub': 'github',
            'lab': 'gitlab',
            'bb': 'bitbucket',
        }
        # setup git command mockup
        self.setup_git_popen()

    def teardown_method(self, method):
        self.log.info('GitRepoMainTestCase.teardown_method({})'.format(method))
        RepositoryService._current = RepositoryMockup(c={})
        self.tempdir.cleanup()

    def setup_args(self, d, args={}):
        cli_args = {
            '--force': False,
            '--help': False,
            '--path': '.',
            '--verbose': 4,
            '--no-clone': False,
            '--tracking': 'master',
            '--alone': False,
            '--add': False,
            '--clone': False,
            '<name>': None,
            '<branch>': None,
            '<target>': self.target,
            '<target_repo>': None,
            '<user>/<repo>': '',
            'add': False,
            'clone': False,
            'create': False,
            'delete': False,
            'fork': False,
            'gist': False,
            'fetch': False,
            'fork': False,
            'list': False,
            'ls': False,
            'open': False,
            '--secret': False,
            '<description>': None,
            '--message': None,
            '<gist>': None,
            '<gist_file>': None,
            '<gist_path>': [],
            'request': False,
            '<request>': None,
            '<local_branch>': None,
            '<remote_branch>': None,
            '<user>/<repo>': None,
        }
        cli_args.update(d)
        cli_args.update(args)
        return cli_args

    def main_add(self, repo, rc=0, args={}):
        if repo:
            create_repo = repo.split('/')[-1]
        else:
            create_repo = 'fubar'
        os.mkdir(os.path.join(self.tempdir.name, create_repo))
        Repo.init(os.path.join(self.tempdir.name, create_repo))
        assert rc == main(self.setup_args({
            'add': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for add".format(rc)
        return RepositoryService._current._did_add

    def main_clone(self, repo, rc=0, args={}):
        assert rc == main(self.setup_args({
            'clone': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for clone".format(rc)
        return RepositoryService._current._did_clone

    def main_create(self, repo=None, rc=0, args={}):
        if repo:
            repo_path = os.path.join(self.tempdir.name, repo.split('/')[-1])
            os.mkdir(repo_path)
            Repo.init(repo_path)
        assert rc == main(self.setup_args({
            'create': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for create".format(rc)
        return RepositoryService._current._did_create

    def main_delete(self, repo=None, rc=0, args={}):
        if repo:
            repo_path = os.path.join(self.tempdir.name, repo.split('/')[-1])
            os.mkdir(repo_path)
            Repo.init(repo_path)
        assert rc == main(self.setup_args({
            'delete': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name,
        }, args)), "Non {} result for delete".format(rc)
        return RepositoryService._current._did_delete

    def main_fork(self, repo=None, rc=0, args={}):
        assert rc == main(self.setup_args({
            'fork': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for fork".format(rc)
        return RepositoryService._current._did_fork

    def main_gist_list(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'list': True,
        }, args)), "Non {} result for gist list".format(rc)
        return RepositoryService._current._did_gist_list

    def main_gist_ls(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'ls': True,
        }, args)), "Non {} result for gist ls".format(rc)
        return RepositoryService._current._did_gist_list

    def main_gist_clone(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'clone': True,
            '--path': self.tempdir.name
        }, args)), "Non {} result for gist clone".format(rc)
        return RepositoryService._current._did_gist_clone

    def main_gist_fetch(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'fetch': True,
        }, args)), "Non {} result for gist fetch".format(rc)
        return RepositoryService._current._did_gist_fetch

    def main_gist_create(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'create': True,
        }, args)), "Non {} result for gist create".format(rc)
        return RepositoryService._current._did_gist_create

    def main_gist_delete(self, rc=0, args={}):
        assert rc == main(self.setup_args({
            'gist': True,
            'delete': True,
        }, args)), "Non {} result for gist delete".format(rc)
        return RepositoryService._current._did_gist_delete

    def main_request_list(self, repo=None, rc=0, args={}):
        assert rc == main(self.setup_args({
            'request': True,
            'list': True,
            '<user>/<repo>': repo,
            '--clone': True,
            '--path': self.tempdir.name
        }, args)), "Non {} result for request list".format(rc)
        return RepositoryService._current._did_request_list

    def main_request_fetch(self, repo=None, rc=0, args={}):
        assert rc == main(self.setup_args({
            'request': True,
            'fetch': True,
            '<user>/<repo>': repo,
            '--clone': True,
            '--path': self.tempdir.name
        }, args)), "Non {} result for request fetch".format(rc)
        return RepositoryService._current._did_request_fetch

    def main_request_create(self, repo=None, rc=0, args={}):
        assert rc == main(self.setup_args({
            'request': True,
            'create': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for request create".format(rc)
        return RepositoryService._current._did_request_create

    def main_open(self, repo=None, rc=0, args={}):
        assert rc == main(self.setup_args({
            'open': True,
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for open".format(rc)
        return RepositoryService._current._did_open

    def main_config(self, target, rc=0, args={}):
        self.target = target
        assert rc == main(self.setup_args({
            'config': True,
            '--config': os.path.join(self.tempdir.name, 'gitconfig')
        }, args)), "Non {} result for config".format(rc)
        with open(os.path.join(self.tempdir.name, 'gitconfig')) as f:
            return f.readlines()

    def main_noop(self, repo, rc=1, args={}):
        assert rc == main(self.setup_args({
            '<user>/<repo>': repo,
            '--path': self.tempdir.name
        }, args)), "Non {} result for no-action".format(rc)


class GitRepoTestCase(TestGitPopenMockupMixin):
    def setup_method(self, method):
        self.log.info('GitRepoTestCase.setup_method({})'.format(method))
        # build temporary directory
        self.tempdir = TemporaryDirectory()
        # when initiating service with no repository, the connection is not triggered
        self.service = self.get_service()
        # setup http api mockup
        self.recorder = betamax.Betamax(self.get_requests_session())
        self.get_requests_session().headers['Accept-Encoding'] = 'identity'
        # setup git command mockup
        self.setup_git_popen()
        # when initiating service with no repository, the connection is not triggered
        self.service.repository = self.repository
        # have git commands logged
        Git.GIT_PYTHON_TRACE = True
        FORMAT = '> %(message)s'
        formatter = logging.Formatter(fmt=FORMAT)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logging.getLogger('git.cmd').removeHandler(logging.NullHandler())
        logging.getLogger('git.cmd').addHandler(handler)
        logging.getLogger('git.cmd').propagate = True
        # have HTTP requests logged
        import http.client
        http.client.HTTPConnection.debuglevel = 1
        logging.getLogger('requests.packages.urllib3').setLevel(logging.DEBUG)
        logging.getLogger('requests.packages.urllib3').propagate = True

    def teardown_method(self, method):
        self.log.info('GitRepoTestCase.teardown_method({})'.format(method))
        self.tempdir.cleanup()

    '''cassette name helper'''

    def _make_cassette_name(self, frame_level=2):
        # returns the name of the function calling the function calling this one
        # in other words, when used in an helper function, returns the name of
        # the test function calling the helper function, to make a cassette name.
        test_function_name = sys._getframe(frame_level).f_code.co_name
        if test_function_name.startswith('test'):
            return '_'.join(['test', self.service.name, test_function_name])
        raise Exception("Helpers functions shall be used only within test functions!")

    '''assertion helpers'''

    def assert_repository_exists(self, namespace, repository):
        try:
            self.service.get_repository(namespace, repository)
        except Exception as err:
            raise AssertionError("Repository {}/{} not found on {}: {}".format(namespace,
                                                                               repository,
                                                                               self.service.name,
                                                                               err)) from err

    def assert_repository_not_exists(self, namespace, repository):
        try:
            self.service.get_repository(namespace, repository)
        except Exception as err:
            return
        #raise AssertionError("Repository {}/{} exists on {}".format(namespace,
        #                                                                       repository,
        #                                                                       self.service.name,
        #                                                                ))

    def assert_added_remote(self, remote):
        try:
            self.repository.remote(remote)
        except ValueError as err:
            raise AssertionError("Remote {} not in repository".format(remote)) from err

    def assert_added_remote_defaults(self):
        self.assert_added_remote(self.service.name)
        self.assert_added_remote('all')

    def assert_tracking_remote(self, remote_name=None, branch_name='master'):
        if not remote_name:
            remote_name = self.service.name
        for branch in self.repository.branches:
            if branch == branch_name:
                assert remote_name in self.repository.branches[0].tracking_branch().name, \
                    'Could not set "{}" as tracking branch master'.format(self.service.name)

    '''test cases templates'''

    def action_fork(self, local_namespace, remote_namespace, repository):
        # hijack subprocess call
        with self.mockup_git(local_namespace, repository):
            # prepare output for git commands
            remote_slug = self.service.format_path(namespace=remote_namespace, repository=repository, rw=True)
            local_slug = self.service.format_path(namespace=local_namespace, repository=repository, rw=True)
            self.set_mock_popen_commands([
                ('git remote add upstream {}'.format(remote_slug), b'', b'', 0),
                ('git remote add all {}'.format(local_slug), b'', b'', 0),
                ('git remote add {} {}'.format(self.service.name, local_slug), b'', b'', 0),
                ('git version', b'git version 2.8.0', b'', 0),
                ('git pull --progress -v {} master'.format(self.service.name), b'', '\n'.join([
                    'POST git-upload-pack (140 bytes)',
                    'remote: Counting objects: 8318, done.',
                    'remote: Compressing objects: 100% (3/3), done.',
                    'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                    'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                    'Resolving deltas: 100% (5126/5126), done.',
                    'From {}:{}/{}'.format(self.service.fqdn, local_namespace, repository),
                    ' * branch            master     -> FETCH_HEAD',
                    ' * [new branch]      master     -> {}/master'.format(self.service.name)]).encode('utf-8'),
                0)
            ])
            with self.recorder.use_cassette(self._make_cassette_name()):
                self.service.connect()
                self.service.fork(remote_namespace, repository)
                # emulate the outcome of the git actions
                self.service.repository.create_remote('upstream', url=remote_slug)
                self.service.repository.create_remote('all', url=local_slug)
                self.service.repository.create_remote(self.service.name, url=local_slug)

    def action_fork__no_clone(self, local_namespace, remote_namespace, repository):
        # hijack subprocess call
        with self.mockup_git(local_namespace, repository):
            # prepare output for git commands
            remote_slug = self.service.format_path(namespace=remote_namespace, repository=repository, rw=True)
            local_slug = self.service.format_path(namespace=local_namespace, repository=repository, rw=True)
            self.set_mock_popen_commands([
                ('git remote add upstream {}'.format(remote_slug), b'', b'', 0),
                ('git remote add all {}'.format(local_slug), b'', b'', 0),
                ('git remote add {} {}'.format(self.service.name, local_slug), b'', b'', 0),
                ('git version', b'git version 2.8.0', b'', 0),
                ('git pull --progress -v {} master'.format(self.service.name), b'', '\n'.join([
                    'POST git-upload-pack (140 bytes)',
                    'remote: Counting objects: 8318, done.',
                    'remote: Compressing objects: 100% (3/3), done.',
                    'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                    'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                    'Resolving deltas: 100% (5126/5126), done.',
                    'From {}:{}/{}'.format(self.service.fqdn, local_namespace, repository),
                    ' * branch            master     -> FETCH_HEAD',
                    ' * [new branch]      master     -> {}/master'.format(self.service.name)]).encode('utf-8'),
                0)
            ])
            with self.recorder.use_cassette(self._make_cassette_name()):
                self.service.connect()
                self.service.fork(remote_namespace, repository)
                # emulate the outcome of the git actions
                self.service.repository.create_remote('upstream', url=remote_slug)
                self.service.repository.create_remote('all', url=local_slug)
                self.service.repository.create_remote(self.service.name, url=local_slug)

    def action_clone(self, namespace, repository):
        # hijack subprocess call
        with self.mockup_git(namespace, repository):
            local_slug = self.service.format_path(namespace=namespace, repository=repository, rw=True)
            self.set_mock_popen_commands([
                ('git init', b'Initialized empty Git repository in /tmp/bar/.git/', b'', 0),
                ('git remote add all {}'.format(local_slug), b'', b'', 0),
                ('git remote add {} {}'.format(self.service.name, local_slug), b'', b'', 0),
                ('git version', b'git version 2.8.0', b'', 0),
                ('git pull --progress -v {} master'.format(self.service.name), b'', '\n'.join([
                    'POST git-upload-pack (140 bytes)',
                    'remote: Counting objects: 8318, done.',
                    'remote: Compressing objects: 100% (3/3), done.',
                    'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                    'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                    'Resolving deltas: 100% (5126/5126), done.',
                    'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                    ' * branch            master     -> FETCH_HEAD',
                    ' * [new branch]      master     -> {}/master'.format(self.service.name)]).encode('utf-8'),
                0)
            ])
            with self.recorder.use_cassette(self._make_cassette_name()):
                self.service.connect()
                self.service.clone(namespace, repository)
                self.service.repository.create_remote('all', url=local_slug)
                self.service.repository.create_remote(self.service.name, url=local_slug)

    def action_create(self, namespace, repository):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            self.service.create(namespace, repository, add=True)
            #
            self.assert_repository_exists(namespace, repository)
            self.assert_added_remote_defaults()

    def action_create__no_add(self, namespace, repository):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            self.service.create(namespace, repository, add=False)
            #
            self.assert_repository_exists(namespace, repository)
            self.assert_added_remote_defaults()

    def action_delete(self, repository, namespace=None):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            if namespace:
                self.service.delete(user=namespace, repo=repository)
            else:
                self.service.delete(repo=repository)
            #
            if not namespace:
                namespace = self.service.user
            self.assert_repository_not_exists(namespace, repository)

    def action_add(self, namespace, repository, alone=False, name=None,
            tracking='master', auto_slug=False, remotes={}):
        with self.recorder.use_cassette(self._make_cassette_name()):
            # init git in the repository's destination
            self.repository.init()
            for remote, url in remotes.items():
                self.repository.create_remote(remote, url)
            self.service.connect()
            self.service.add(user=namespace, repo=repository, alone=alone, name=name, tracking=tracking, auto_slug=auto_slug)
            #
            if not tracking:
                if not alone and not name:
                    self.assert_added_remote_defaults()
                elif not alone and name:
                    self.assert_added_remote(name)
                    self.assert_added_remote('all')
                elif alone and not name:
                    self.assert_added_remote(self.service.name)
                elif alone and name:
                    self.assert_added_remote(name)
            else:
                if not alone and not name:
                    self.assert_added_remote_defaults()
                    self.assert_tracking_remote()
                elif not alone and name:
                    self.assert_added_remote(name)
                    self.assert_added_remote('all')
                    self.assert_tracking_remote(name)
                elif alone and not name:
                    self.assert_added_remote(self.service.name)
                    self.assert_tracking_remote(branch_name=tracking)
                elif alone and name:
                    self.assert_added_remote(name)
                    self.assert_tracking_remote(name, tracking)

    def action_list(self, namespace, _long=False):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            return list(self.service.list(namespace, _long=_long))

    def action_request_list(self, namespace, repository, rq_list_data=[]):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            requests = list(self.service.request_list(user=namespace, repo=repository))
            for i, rq in enumerate(rq_list_data):
                assert requests[i] == rq

    def action_request_fetch(self, namespace, repository, request, pull=False, fail=False, remote_branch='pull', local_branch='requests'):
        local_slug = self.service.format_path(namespace=namespace, repository=repository, rw=False)
        with self.recorder.use_cassette(self._make_cassette_name()):
            with self.mockup_git(namespace, repository):
                self.set_mock_popen_commands([
                    ('git remote add all {}'.format(local_slug), b'', b'', 0),
                    ('git remote add {} {}'.format(self.service.name, local_slug), b'', b'', 0),
                    ('git remote get-url --all all', local_slug.encode('utf-8'), b'', 0),
                    ('git remote get-url --all {}'.format(self.service.name), local_slug.encode('utf-8'), b'', 0),
                    ('git version', b'git version 2.8.0', b'', 0),
                    ('git pull --progress -v {} master'.format(self.service.name), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * branch            master     -> FETCH_HEAD',
                        ' * [new branch]      master     -> {1}/{0}'.format(request, local_branch)]).encode('utf-8'),
                    0),
                    ('git version', b'git version 2.8.0', b'', 0),
                    ('git fetch --progress -v {0} {2}/{1}/head:{3}/{1}'.format(
                            self.service.name,
                            request,
                            remote_branch,
                            local_branch), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * [new branch]      master     -> request/{}'.format(request)]).encode('utf-8'),
                    0)
                ])
                self.service.connect()
                self.service.clone(namespace, repository, rw=False)
            if not fail:
                self.service.repository.create_remote('all', url=local_slug)
                self.service.repository.create_remote(self.service.name, url=local_slug)
            with self.mockup_git(namespace, repository):
                self.set_mock_popen_commands([
                    ('git version', b'git version 2.8.0', b'', 0),
                    ('git fetch --progress --update-head-ok -v {0} {2}/{1}/head:{3}/{1}'.format(
                            self.service.name,
                            request,
                            remote_branch,
                            local_branch), b'', '\n'.join([
                        'POST git-upload-pack (140 bytes)',
                        'remote: Counting objects: 8318, done.',
                        'remote: Compressing objects: 100% (3/3), done.',
                        'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                        'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                        'Resolving deltas: 100% (5126/5126), done.',
                        'From {}:{}/{}'.format(self.service.fqdn, namespace, repository),
                        ' * [new branch]      master     -> {1}/{0}'.format(request, local_branch)]).encode('utf-8'),
                    0)
                ])
                self.service.request_fetch(repo=repository, user=namespace, request=request)

    def action_request_create(self,
            namespace, repository,
            title, description,
            source_branch='pr-test',
            target_branch='master',
            create_repository='test_create_requests',
            auto_slug=False):
        '''
        Here we are testing the subcommand 'request create'.

        this test needs sensibly more preparation than other tests, because to create
        a pull request, you need:

        * a repository with commits on both the service and your workspace
        * a new branch with new commits, that has been pushed on the service

        So that's what we're doing below:
            * create a test project on the service,
            * populate the temporary git repository with it
            * create a commit and push it to the service as master
            * create a branch in the workspace
            * create a commit and push it to the service as pr-test

        Then we test the feature:
            * using the branch create a pull request and check the pull request is there

        Finally clean the remote repository

        So all the contextual work is only done
        '''
        cassette_name = self._make_cassette_name()
        will_record = 'never' != self.recorder.config.default_cassette_options['record_mode'] \
                and not os.path.exists(os.path.join(self.recorder.config.cassette_library_dir, cassette_name+'.json'))

        @contextmanager
        def prepare_project_for_test():
            if will_record:
                self.service.connect()
                # let's create a project and add it to current repository
                self.service.create(namespace, create_repository, add=True)
                # make a modification, commit and push it
            with open(os.path.join(self.repository.working_dir, 'first_file'), 'w') as test:
                test.write('he who makes a beast of himself gets rid of the pain of being a man. Dr Johnson')
            self.repository.git.config('user.name', 'travis')
            self.repository.git.config('user.email', 'travis@fake.host')
            self.repository.git.add('first_file')
            self.repository.git.commit(message='First commit')
            if will_record:
                self.repository.git.push(self.service.name, 'master')
                # create a new branch
                new_branch = self.repository.create_head(source_branch, 'HEAD')
                self.repository.head.reference = new_branch
                self.repository.head.reset(index=True, working_tree=True)
                # make a modification, commit and push it to that branch
                with open(os.path.join(self.repository.working_dir, 'second_file'), 'w') as test:
                    test.write('La meilleure façon de ne pas avancer est de suivre une idée fixe. J.Prévert')
                self.repository.git.add('second_file')
                self.repository.git.commit(message='Second commit')
                self.repository.git.push(service, source_branch)
            else:
                import git
                self.service._extracts_ref = lambda *a: git.Reference(
                        self.service.repository,
                        '{}/{}'.format(namespace, repository),
                        check_path=False)
                self.service.repository.head.reference = git.Head(self.service.repository, path='refs/heads/pr-test')
            existing_remotes = [r.name for r in self.repository.remotes]
            if 'all' in existing_remotes:
                r_all = self.repository.remote('all')
            else:
                r_all = self.repository.create_remote('all', '')
            for name in ('github', 'gitlab', 'bitbucket', 'gogs', 'upstream'):
                if name not in existing_remotes:
                    kw = dict(user=namespace, project=repository, host=name)
                    self.repository.create_remote(name, 'git@{host}.com:{user}/{project}'.format(**kw))
                    r_all.add_url('git@{host}.com:{user}/{project}'.format(**kw))
            yield
            if will_record:
                self.service.delete(create_repository)

        #self.service.repository = self.repository

        def edit_stub(repository, branch):
            return 'title', 'description'

        with prepare_project_for_test():
            with self.recorder.use_cassette(cassette_name):
                self.service.connect()
                def test_edit(repository, from_branch):
                    return "PR title", "PR body"
                request = self.service.request_create(
                    onto_user=namespace,
                    onto_repo=repository,
                    from_branch=source_branch,
                    onto_branch=target_branch,
                    title=title,
                    description=description,
                    auto_slug=auto_slug,
                    edit=test_edit)
                return request

    def action_gist_list(self, gist=None, gist_list_data=[]):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            if gist is None:
                gists = list(self.service.gist_list())
                for i, g in enumerate(gist_list_data):
                    assert gists[i] == g
            else:
                gist_files = list(self.service.gist_list(gist))
                for i, gf in enumerate(gist_list_data):
                    assert gist_files[i] == gf

    def action_gist_clone(self, gist, clone_url=None):
        with self.mockup_git(None, None):
            self.set_mock_popen_commands([
                ('git version', b'git version 2.8.0', b'', 0),
                ('git remote add gist {}.git'.format(clone_url or gist), b'', b'', 0),
                ('git pull --progress -v gist master', b'', b'\n'.join([
                    b'POST git-upload-pack (140 bytes)',
                    b'remote: Counting objects: 8318, done.',
                    b'remote: Compressing objects: 100% (3/3), done.',
                    b'remote: Total 8318 (delta 0), reused 0 (delta 0), pack-reused 8315',
                    b'Receiving objects: 100% (8318/8318), 3.59 MiB | 974.00 KiB/s, done.',
                    b'Resolving deltas: 100% (5126/5126), done.',
                    bytes('From {}'.format(clone_url or gist), 'utf-8'),
                    b' * branch            master     -> FETCH_HEAD']),
                0),
            ])
            with self.recorder.use_cassette(self._make_cassette_name()):
                self.service.connect()
                self.service.gist_clone(gist)


    def action_gist_fetch(self, gist, gist_file=None):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            content = self.service.gist_fetch(gist, gist_file)
            return content

    def action_gist_create(self, description, gist_files, secret):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            content = self.service.gist_create(gist_files, description, secret)

    def action_gist_delete(self, gist):
        with self.recorder.use_cassette(self._make_cassette_name()):
            self.service.connect()
            content = self.service.gist_delete(gist)

    def action_open(self, namespace, repository):
        self.set_mock_popen_commands([
            ('xdg-open {}'.format(self.service.format_path(namespace=namespace, repository=repository)), b'', b'', 0),
            ('open {}'.format(self.service.format_path(namespace=namespace, repository=repository)), b'', b'', 0),
        ])
        with Replace('subprocess.Popen', self.Popen):
            self.service.open(user=namespace, repo=repository)




#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.github')

from ..service import register_target, RepositoryService, os
from ...exceptions import ResourceError, ResourceExistsError, ResourceNotFoundError, ArgumentError
from ...tools import columnize

import github3

from git.exc import GitCommandError

from datetime import datetime

GITHUB_COM_FQDN = 'github.com'

@register_target('hub', 'github')
class GithubService(RepositoryService):
    fqdn = GITHUB_COM_FQDN

    def __init__(self, *args, **kwarg):
        self.gh = github3.GitHub()
        super(GithubService, self).__init__(*args, **kwarg)

    def connect(self):
        if self.fqdn != GITHUB_COM_FQDN:
            # upgrade self.gh from a GitHub object to a GitHubEnterprise object
            gh = github3.GitHubEnterprise(RepositoryService.build_url(self))
            self.gh._session.base_url = gh._session.base_url
            gh._session = self.gh._session
            self.gh = gh
            # propagate ssl certificate parameter
            self.gh._session.verify = self.session_certificate or not self.session_insecure
        try:
            self.gh.login(token=self._privatekey)
            self.username = self.gh.user().login
        except github3.models.GitHubError as err:
            if 401 == err.code:
                if not self._privatekey:
                    raise ConnectionError('Could not connect to Github. '
                                          'Please configure .gitconfig '
                                          'with your github private key.') from err
                else:
                    raise ConnectionError('Could not connect to Github. '
                                          'Check your configuration and try again.') from err

    def create(self, user, repo, add=False):
        try:
            if user != self.username:
                org = self.gh.organization(user)
                if org:
                    org.create_repo(repo)
                else:
                    raise ResourceNotFoundError("Namespace {} neither an organization or current user.".format(user))
            else:
                self.gh.create_repo(repo)
        except github3.models.GitHubError as err:
            if err.code == 422 or err.message == 'name already exists on this account':
                raise ResourceExistsError("Project already exists.") from err
            else: # pragma: no cover
                raise ResourceError("Unhandled error.") from err
        if add:
            self.add(user=user, repo=repo, tracking=self.name)

    def fork(self, user, repo):
        try:
            return self.gh.repository(user, repo).create_fork().full_name
        except github3.models.GitHubError as err:
            if err.message == 'name already exists on this account':
                raise ResourceExistsError("Project already exists.") from err
            else: # pragma: no cover
                raise ResourceError("Unhandled error: {}".format(err)) from err

    def delete(self, repo, user=None):
        if not user:
            user = self.username
        try:
            repository = self.gh.repository(user, repo)
            if repository:
                result = repository.delete()
            if not repository or not result:
                raise ResourceNotFoundError('Cannot delete: repository {}/{} does not exists.'.format(user, repo))
        except github3.models.GitHubError as err: # pragma: no cover
            if err.code == 403:
                raise ResourcePermissionError('You don\'t have enough permissions for deleting the repository. '
                                              'Check the namespace or the private token\'s privileges') from err
            raise ResourceError('Unhandled exception: {}'.format(err)) from err

    def list(self, user, _long=False):
        if not self.gh.user(user):
            raise ResourceNotFoundError("User {} does not exists.".format(user))

        repositories = self.gh.iter_user_repos(user)
        if not _long:
            repositories = list(["/".join([user, repo.name]) for repo in repositories])
            yield "{}"
            yield ("Total repositories: {}".format(len(repositories)),)
            yield from columnize(repositories)
        else:
            yield "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}"
            yield ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif', 'Name']
            for repo in repositories:
                try:
                    if repo.updated_at.year < datetime.now().year:
                        date_fmt = "%b %d %Y"
                    else:
                        date_fmt = "%b %d %H:%M"

                    status = ''.join([
                        'F' if repo.fork else ' ',               # is a fork?
                        'P' if repo.private else ' ',            # is private?
                    ])
                    nb_pulls = len(list(repo.iter_pulls()))
                    nb_issues = len(list(repo.iter_issues())) - nb_pulls
                    yield [
                        # status
                        status,
                        # stats
                        str(len(list(repo.iter_commits()))),          # number of commits
                        str(nb_pulls),                                # number of pulls
                        str(nb_issues),                               # number of issues
                        str(repo.forks),                              # number of forks
                        str(len(list(repo.iter_contributors()))),     # number of contributors
                        str(repo.watchers),                           # number of subscribers
                        str(repo.stargazers or 0),                    # number of ♥
                        # info
                        repo.language or '?',                      # language
                        repo.updated_at.strftime(date_fmt),      # date
                        '/'.join([user, repo.name]),             # name
                    ]
                except Exception as err:
                    if 'Git Repository is empty.' == err.args[0].json()['message']:
                        yield [
                            # status
                            'E',
                            # stats
                            'ø',     # number of commits
                            'ø',     # number of pulls
                            'ø',     # number of issues
                            'ø',     # number of forks
                            'ø',     # number of contributors
                            'ø',     # number of subscribers
                            'ø',     # number of ♥
                            # info
                            '?',     # language
                            repo.updated_at.strftime(date_fmt),      # date
                            '/'.join([user, repo.name]),             # name
                        ]
                    else:
                        print("Cannot show repository {}: {}".format('/'.join([user, repo.name]), err))

    def _format_gist(self, gist):
        return gist.split('https://gist.github.com/')[-1].split('.git')[0]

    def gist_list(self, gist=None):
        if not gist:
            yield "{:45.45} {}"
            yield 'title', 'url'
            for gist in self.gh.iter_gists(self.gh.user().login):
                yield gist.description, gist.html_url
        else:
            gist = self.gh.gist(self._format_gist(gist))
            if gist is None:
                raise ResourceNotFoundError('Gist does not exists.')
            yield "{:15}\t{:7}\t{}"
            yield 'language', 'size', 'name'
            for gist_file in gist.iter_files():
                yield (gist_file.language if gist_file.language else 'Raw text',
                        gist_file.size,
                        gist_file.filename)


    def gist_fetch(self, gist, fname=None):
        try:
            gist = self.gh.gist(self._format_gist(gist))
        except Exception as err:
            raise ResourceNotFoundError('Error while fetching gist') from err
        if not gist:
            raise ResourceNotFoundError('Could not find gist')
        if gist.files == 1 and not fname:
            gist_file = list(gist.iter_files())[0]
        else:
            for gist_file in gist.iter_files():
                if gist_file.filename == fname:
                    break
            else:
                raise ResourceNotFoundError('Could not find file within gist.')

        return gist_file.content

    def gist_clone(self, gist):
        try:
            gist = self.gh.gist(gist.split('https://gist.github.com/')[-1])
        except Exception as err:
            raise ResourceNotFoundError('Could not find gist') from err
        remote = self.repository.create_remote('gist', gist.git_push_url)
        self.pull(remote, 'master')

    def gist_create(self, gist_pathes, description, secret=False):
        def load_file(fname, path='.'):
            with open(os.path.join(path, fname), 'r') as f:
                return {'content': f.read()}

        gist_files = dict()
        for gist_path in gist_pathes:
            if not os.path.isdir(gist_path):
                gist_files[os.path.basename(gist_path)] = load_file(gist_path)
            else:
                for gist_file in os.listdir(gist_path):
                    if not os.path.isdir(os.path.join(gist_path, gist_file)) and not gist_file.startswith('.'):
                        gist_files[gist_file] = load_file(gist_file, gist_path)

        gist = self.gh.create_gist(
                description=description,
                files=gist_files,
                public=not secret # isn't it obvious? ☺
            )

        return gist.html_url

    def gist_delete(self, gist_id):
        gist = self.gh.gist(self._format_gist(gist_id))
        if not gist:
            raise ResourceNotFoundError('Could not find gist')
        gist.delete()

    def request_create(self, onto_user, onto_repo, from_branch, onto_branch, title=None, description=None, auto_slug=False, edit=None):
        onto_project = self.gh.repository(onto_user, onto_repo)

        if not onto_project:
            raise ResourceNotFoundError('Could not find project `{}/{}`!'.format(onto_user, onto_repo))

        from_reposlug = self.guess_repo_slug(self.repository, self, resolve_targets=['{service}'])
        if from_reposlug:
            from_user, from_repo = from_reposlug.split('/')
            if (onto_user, onto_repo) == (from_user, from_repo):
                from_project = onto_project
            else:
                from_project = self.gh.repository(from_user, from_repo)
        else:
            from_project = None

        if not from_project:
            raise ResourceNotFoundError('Could not find project `{}`!'.format(from_user, from_repo))

        # when no repo slug has been given to `git-repo X request create`
        # then chances are current project is a fork of the target
        # project we want to push to
        if auto_slug and onto_project.fork:
            onto_user = onto_project.parent.owner.login
            onto_repo = onto_project.parent.name
            onto_project = self.gh.repository(onto_user, onto_repo)

        # if no onto branch has been defined, take the default one
        # with a fallback on master
        if not from_branch:
            from_branch = self.repository.active_branch.name

        # if no from branch has been defined, chances are we want to push
        # the branch we're currently working on
        if not onto_branch:
            onto_branch = self.get_project_default_branch(onto_project)

        from_target = '{}:{}'.format(from_user, from_branch)
        onto_target = '{}/{}:{}'.format(onto_user, onto_project, onto_branch)

        # translate from github username to git remote name
        if not title and not description and edit:
            title, description = edit(self.repository, from_branch, onto_target)
            if not title and not description:
                raise ArgumentError('Missing message for request creation')

        try:
            request = onto_project.create_pull(title,
                    head=from_target,
                    base=onto_branch,
                    body=description)

            return {
                'local': from_branch,
                'project': '/'.join([onto_user, onto_repo]),
                'remote': onto_branch,
                'ref': request.number,
                'url': request.html_url
            }

        except github3.models.GitHubError as err:
            if err.code == 422:
                if err.message == 'Validation Failed':
                    for error in err.errors:
                        if 'message' in error:
                            raise ResourceError(error['message'])
                        if error.get('code', '') == 'invalid':
                            if error.get('field', '') == 'head':
                                raise ResourceError(
                                        'Invalid source branch. ' \
                                        'Check it has been pushed first.')
                            if error.get('field', '') == 'base':
                                raise ResourceError( 'Invalid target branch.')
                    raise ResourceError("Unhandled formatting error: {}".format(err.errors))
            raise ResourceError(err.message)

    def request_list(self, user, repo):
        repository = self.gh.repository(user, repo)
        yield "{}\t{:<60}\t{}"
        yield 'id', 'title', 'URL'
        for pull in repository.iter_pulls():
            yield str(pull.number), pull.title, pull.links['html']

    def request_fetch(self, user, repo, request, pull=False, force=False):
        if pull:
            raise NotImplementedError('Pull operation on requests for merge are not yet supported')
        try:
            remote_names = list(self._convert_user_into_remote(user))
            for remote in self.repository.remotes:
                if remote.name in remote_names:
                    local_branch_name = 'requests/{}/{}'.format(self.name,request)
                    self.fetch(
                        remote,
                        'pull/{}/head'.format(request),
                        local_branch_name,
                        force=force
                    )
                    return local_branch_name
            else:
                raise ResourceNotFoundError('Could not find remote {}'.format(self.name))
        except GitCommandError as err:
            if 'Error when fetching: fatal: Couldn\'t find remote ref' in err.command[0]:
                raise ResourceNotFoundError('Could not find opened request #{}'.format(request)) from err
            raise err

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        import platform
        if cls.fqdn != GITHUB_COM_FQDN:
            gh = github3.GitHubEnterprise()
        else:
            gh = github3.GitHub()
        gh.login(login, password, two_factor_callback=lambda: prompt('2FA code> '))
        try:
            auth = gh.authorize(login, password,
                    scopes=[ 'repo', 'delete_repo', 'gist' ],
                    note='git-repo2 token used on {}'.format(platform.node()),
                    note_url='https://github.com/guyzmo/git-repo')
            return auth.token
        except github3.models.GitHubError as err:
            if len(err.args) > 0 and 422 == err.args[0].status_code:
                raise ResourceExistsError("A token already exist for this machine on your github account.")
            else:
                raise err

    def get_parent_project_url(self, user, project, rw=True):
        parent = self.gh.repository(user, project).parent
        if not parent:
            return None
        return self.format_path(
                repository=parent.name,
                namespace=parent.owner.login,
                rw=rw)

    @property
    def user(self):
        return self.gh.user().login

    def get_repository(self, user, repo):
        repository = self.gh.repository(user, repo)
        if not repository:
            raise ResourceNotFoundError('Cannot delete: repository {}/{} does not exists.'.format(user, repo))
        return repository

    @staticmethod
    def is_repository_empty(project):
        return project.size == 0

    @staticmethod
    def get_project_default_branch(project):
       return project.default_branch or 'master'

